#include <stdint.h>
#include <string.h>
#include "nordic_common.h"
#include "nrf.h"
#include "app_error.h"
#include "nrf_gpio.h"
#include "nrf51_bitfields.h"
#include "ble.h"
#include "ble_hci.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_conn_params.h"
#include "app_scheduler.h"
#include "softdevice_handler.h"
#include "app_timer.h"
#include "app_timer_appsh.h"
#include "ble_error_log.h"
#include "app_gpiote.h"
#include "app_button.h"
#include "ble_debug_assert_handler.h"
#include "pstorage.h"
#include "twi_master.h"
#include "nrf_delay.h"
#include "pstorage.h"
#include "ble_dfu.h"
#include "dfu_app_handler.h"
#include "math.h"

#define IS_SRVC_CHANGED_CHARACT_PRESENT 1                                           /**< Include or not the service_changed characteristic. if not enabled, the server's database cannot be changed for the lifetime of the device*/

#define WAKEUP_BUTTON_PIN               INT_ACC                                     /**< INT2 (Accelerometer) used to wake up the application. */
#define DEVICE_NAME                     "SECURITE"                                   /**< Name of device. Will be included in the advertising data. 9 characters maximum */
#define LOCAL_NAME                      "Warm sole"
#define DEVICE_NAME_LEFT                "Left sole"
#define LOCAL_NAME_LEFT                 "Warm L"                                    /**< Name of device. Will be included in the advertising data. */
#define DEVICE_NAME_RIGHT               "Right sole"
#define LOCAL_NAME_RIGHT                "Warm R"                                    /**< Name of device. Will be included in the advertising data. */
#define MANUFACTURER_NAME               "GlaglaShoes"                               /**< Manufacturer. Will be passed to Device Information Service. */

#define APP_ADV_INTERVAL                1500                                        /**< The advertising interval (in units of 0.625 ms. This value corresponds to 40 ms). */
#define APP_ADV_TIMEOUT_IN_SECONDS      BLE_GAP_ADV_TIMEOUT_GENERAL_UNLIMITED       /**< The advertising timeout (in units of seconds). */
#define APP_ADV_TIMEOUT                 300                                         /**< The advertising timeout (in units of seconds) managed by Timer_1s. */

#define MIN_CONN_INTERVAL               MSEC_TO_UNITS(100, UNIT_1_25_MS)            /**< Minimum acceptable connection interval (0.5 seconds). */
#define MAX_CONN_INTERVAL               MSEC_TO_UNITS(200, UNIT_1_25_MS)            /**< Maximum acceptable connection interval (1 second). */
#define SLAVE_LATENCY                   0                                           /**< Slave latency. */
#define CONN_SUP_TIMEOUT                MSEC_TO_UNITS(10000, UNIT_10_MS)            /**< Connection supervisory timeout (10 seconds). */
#define FIRST_CONN_PARAMS_UPDATE_DELAY  APP_TIMER_TICKS(5000, APP_TIMER_PRESCALER)  /**< Time from initiating event (connect or start of notification) to first time sd_ble_gap_conn_param_update is called (5 seconds). */
#define NEXT_CONN_PARAMS_UPDATE_DELAY   APP_TIMER_TICKS(30000, APP_TIMER_PRESCALER) /**< Time between each call to sd_ble_gap_conn_param_update after the first call (30 seconds). */
#define MAX_CONN_PARAMS_UPDATE_COUNT    3                                           /**< Number of attempts before giving up the connection parameter negotiation. */
#define SECURITY_REQUEST_DELAY         APP_TIMER_TICKS(4000, APP_TIMER_PRESCALER)

APP_TIMER_DEF(timer_1s_id);                                                         /**< timer 1s for ADC reads */
#define TIMER_1S_INTERVAL               APP_TIMER_TICKS(1000, APP_TIMER_PRESCALER)  /**< timer 1s interval (ticks). */

APP_TIMER_DEF(m_sec_req_timer_id);
APP_TIMER_DEF(timer_accelerometer_id);                                              /**< timer 25ms for accelerometer*/
#define TIMER_ACC_INTERVAL              APP_TIMER_TICKS(25, APP_TIMER_PRESCALER)    /**< timer 25ms interval (ticks). */

#define APP_GPIOTE_MAX_USERS            2                                           /**< Maximum number of users of the GPIOTE handler. */

#define SEC_PARAM_TIMEOUT               30                                          /**< Timeout for Pairing Request or Security Request (in seconds). */
#define SEC_PARAM_BOND                  1                                           /**< Perform bonding. */
#define SEC_PARAM_MITM                  0                                           /**< Man In The Middle protection not required. */
#define SEC_PARAM_IO_CAPABILITIES       BLE_GAP_IO_CAPS_NONE                        /**< No I/O capabilities. */
#define SEC_PARAM_OOB                   0                                           /**< Out Of Band data not available. */
#define SEC_PARAM_MIN_KEY_SIZE          7                                           /**< Minimum encryption key size. */
#define SEC_PARAM_MAX_KEY_SIZE          16                                          /**< Maximum encryption key size. */

pstorage_handle_t                       flash_handler;                               /**< Handle of pstorage block */

#define SCHED_MAX_EVENT_DATA_SIZE       sizeof(app_timer_event_t)                   /**< Maximum size of scheduler events. Note that scheduler BLE stack events do not contain any data, as the events are being pulled from the stack in the event handler. */
#define SCHED_QUEUE_SIZE                20                                          /**< Maximum number of events in the scheduler queue. */


#ifdef BLE_DFU_APP_SUPPORT
#define DFU_REV_MAJOR                    0x00                                       /** DFU Major revision number to be exposed. */
#define DFU_REV_MINOR                    0x01                                       /** DFU Minor revision number to be exposed. */
#define DFU_REVISION                     ((DFU_REV_MAJOR << 8) | DFU_REV_MINOR)     /** DFU Revision number to be exposed. Combined of major and minor versions. */
#define APP_SERVICE_HANDLE_START         0x000C                                     /**< Handle of first application specific service when when service changed characteristic is present. */
#define BLE_HANDLE_MAX                   0xFFFF                                     /**< Max handle value in BLE. */

STATIC_ASSERT(IS_SRVC_CHANGED_CHARACT_PRESENT);                                     /** When having DFU Service support in application the Service Changed Characteristic should always be present. */
#endif // BLE_DFU_APP_SUPPORT

#define APP_TIMER_PRESCALER             16                      /**< Value of the RTC1 PRESCALER register. */
#define APP_TIMER_MAX_TIMERS            10                       /**< Maximum number of simultaneously created timers. */
#define APP_TIMER_OP_QUEUE_SIZE         10                      /**< Size of timer operation queues. */
//Global variable
 

uint16_t        Time_Production_Autotest = 0;
uint16_t        battery_mv;
ble_gap_addr_t  p_ble_address;

static ble_gap_sec_params_t             m_sec_params;                               /**< Security requirements for this application. */
#ifdef BLE_DFU_APP_SUPPORT
    static ble_dfu_t                        m_dfus;                                 /**< Structure used to identify the DFU service. */
#endif // BLE_DFU_APP_SUPPORT


#define READ_ACC                                            0x0008
#define STOP_ADV                                            0x0080
#define RESTART_ADV                                         0x0100
#define POWER_OFF                                           0x8000
#define NOTHING                                             0x0000


enum SYSTEM_STATUS_E
{
    BLE_OFF,
    BLE_ADV,
    BLE_CONNECTED,
    STOP,
};

typedef struct status_s
{
    uint16_t                                            jobtodo;
    uint16_t                                            m_conn_handle;
    bool                                                system;
    enum SYSTEM_STATUS_E                                system_status;
} status_t;
static status_t status;

void pstorage_sys_event_handler(uint32_t p_evt);


void app_error_handler(uint32_t error_code, uint32_t line_num, const uint8_t * p_file_name)
{
}


void assert_nrf_callback(uint16_t line_num, const uint8_t * p_file_name)
{
}

void flash_callback_handler(pstorage_handle_t * p_handle,
    uint8_t             op_code,
    uint32_t            result,
    uint8_t           * p_data,
    uint32_t            data_len)
{
    switch (op_code) {
    case PSTORAGE_STORE_OP_CODE:
        if (result == NRF_SUCCESS) {
            ;                           // Nothing to do : Store operation successful.
        }
        else {
            APP_ERROR_CHECK(result);    // Store operation failed.
        }
        break;
    case PSTORAGE_LOAD_OP_CODE:
        if (result == NRF_SUCCESS) {
            ;                           // Nothing to do : Store operation successful.
        }
        else {
            APP_ERROR_CHECK(result);    // Store operation failed.
        }
        break;
    case PSTORAGE_CLEAR_OP_CODE:
        if (result == NRF_SUCCESS) {
            ;                           // Nothing to do : Store operation successful.
        }
        else {
            APP_ERROR_CHECK(result);    // Store operation failed.
        }
        break;
    case PSTORAGE_UPDATE_OP_CODE:
        if (result == NRF_SUCCESS) {
            ;                           // Nothing to do : Store operation successful.
        }
        else {
            APP_ERROR_CHECK(result);    // Store operation failed.
        }
        break;
    default:
        break;
    }
}


static void timers_init(void)
{
    uint32_t    err_code;
    // Initialize timer module, making it use the scheduler
    APP_TIMER_APPSH_INIT(APP_TIMER_PRESCALER, APP_TIMER_OP_QUEUE_SIZE, true);
}


static void gap_params_init(void)
{
    uint32_t                err_code;
    ble_gap_conn_params_t   gap_conn_params;
    ble_gap_conn_sec_mode_t sec_mode;

    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
        (const uint8_t *)DEVICE_NAME,
        strlen(DEVICE_NAME));

    APP_ERROR_CHECK(err_code);

    /* appearance value matching the application's use case.*/
    err_code = sd_ble_gap_appearance_set(BLE_APPEARANCE_GENERIC_TAG);
    APP_ERROR_CHECK(err_code);

    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MIN_CONN_INTERVAL;
    gap_conn_params.max_conn_interval = MAX_CONN_INTERVAL;
    gap_conn_params.slave_latency = SLAVE_LATENCY;
    gap_conn_params.conn_sup_timeout = CONN_SUP_TIMEOUT;

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);

    err_code = sd_ble_gap_tx_power_set(RADIO_TXPOWER_TXPOWER_Pos4dBm);
    APP_ERROR_CHECK(err_code);
}


static void advertising_init(uint8_t adv_flags)
{
    uint32_t                    err_code;
    ble_advdata_t               advdata;
    ble_advdata_t               scanrsp;
    ble_advdata_manuf_data_t    manufacturer;
    uint8_array_t               adv_manuf_data_array;
    uint8_t                     adv_manuf_data_data[7];

    // Use UUIDs for service(s) used in your application.
    ble_uuid_t adv_uuids[] = { { 0xCDA5, BLE_UUID_TYPE_BLE } };

    // Build and set advertising data
    memset(&advdata, 0, sizeof(advdata));

    advdata.name_type = BLE_ADVDATA_FULL_NAME;
    advdata.include_appearance = true;
    advdata.flags = adv_flags;

    adv_manuf_data_array.p_data = adv_manuf_data_data;
    adv_manuf_data_array.size = 7;

    manufacturer.company_identifier = 7;
    manufacturer.data = adv_manuf_data_array;

    advdata.p_manuf_specific_data = &manufacturer;

    memset(&scanrsp, 0, sizeof(scanrsp));

    scanrsp.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]);
    scanrsp.uuids_complete.p_uuids = adv_uuids;

    err_code = ble_advdata_set(&advdata, NULL);
    APP_ERROR_CHECK(err_code);
}


static void advertising_stop(void)
{
    uint32_t err_code;

    err_code = sd_ble_gap_adv_stop();
    APP_ERROR_CHECK(err_code);
}


#ifdef BLE_DFU_APP_SUPPORT
/** @snippet [DFU BLE Reset prepare] */
static void reset_prepare(void)
{
    uint32_t err_code;

    NRF_WDT->RR[0] = WDT_RR_RR_Reload;

    if (status.m_conn_handle != BLE_CONN_HANDLE_INVALID) {
        // Disconnect from peer.
        err_code = sd_ble_gap_disconnect(status.m_conn_handle, BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION);
        APP_ERROR_CHECK(err_code);
    }
    else {
        // If not connected, then the device will be advertising. Hence stop the advertising.
        advertising_stop();
    }

    err_code = ble_conn_params_stop();
    APP_ERROR_CHECK(err_code);
}

#endif // BLE_DFU_APP_SUPPORT


static void services_init(void)
{
    uint32_t                    err_code;

#ifdef BLE_DFU_APP_SUPPORT
    /** @snippet [DFU BLE Service initialization] */
    ble_dfu_init_t   dfus_init;

    // Initialize the Device Firmware Update Service.
    memset(&dfus_init, 0, sizeof(dfus_init));

    dfus_init.evt_handler = dfu_app_on_dfu_evt;
    dfus_init.error_handler = NULL;
    dfus_init.evt_handler = dfu_app_on_dfu_evt;
    dfus_init.revision = DFU_REVISION;

    err_code = ble_dfu_init(&m_dfus, &dfus_init);
    APP_ERROR_CHECK(err_code);

    dfu_app_reset_prepare_set(reset_prepare);

#endif // BLE_DFU_APP_SUPPORT
}


static void sec_params_init(void)
{
    m_sec_params.bond = SEC_PARAM_BOND;
    m_sec_params.mitm = SEC_PARAM_MITM;
    m_sec_params.io_caps = SEC_PARAM_IO_CAPABILITIES;
    m_sec_params.oob = SEC_PARAM_OOB;
    m_sec_params.min_key_size = SEC_PARAM_MIN_KEY_SIZE;
    m_sec_params.max_key_size = SEC_PARAM_MAX_KEY_SIZE;
}


static void on_conn_params_evt(ble_conn_params_evt_t * p_evt)
{
    uint32_t err_code;

    if (p_evt->evt_type == BLE_CONN_PARAMS_EVT_FAILED) {
        err_code = sd_ble_gap_disconnect(status.m_conn_handle, BLE_HCI_CONN_INTERVAL_UNACCEPTABLE);
        APP_ERROR_CHECK(err_code);
    }
}


static void conn_params_error_handler(uint32_t nrf_error)
{
    APP_ERROR_HANDLER(nrf_error);
}


static void conn_params_init(void)
{
    uint32_t               err_code;
    ble_conn_params_init_t cp_init;
    memset(&cp_init, 0, sizeof(cp_init));
    cp_init.p_conn_params = NULL;
    cp_init.first_conn_params_update_delay = FIRST_CONN_PARAMS_UPDATE_DELAY;
    cp_init.next_conn_params_update_delay = NEXT_CONN_PARAMS_UPDATE_DELAY;
    cp_init.max_conn_params_update_count = MAX_CONN_PARAMS_UPDATE_COUNT;
    cp_init.start_on_notify_cccd_handle = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail = false;
    cp_init.evt_handler = on_conn_params_evt;
    cp_init.error_handler = conn_params_error_handler;
    err_code = ble_conn_params_init(&cp_init);
    APP_ERROR_CHECK(err_code);
}


static void advertising_start(void)
{
    uint32_t             err_code;
    ble_gap_adv_params_t adv_params;
    ble_gap_whitelist_t  whitelist;

    // Initialize advertising parameters with defaults values
    memset(&adv_params, 0, sizeof(adv_params));
    memset(&whitelist, 0, sizeof(whitelist));

    ble_gap_addr_t temp2;

    //memset(&temp2, 0, sizeof(temp2));
    temp2.addr[5] = 0x55;
    temp2.addr[4] = 0xe3;
    temp2.addr[3] = 0xb1;
    temp2.addr[2] = 0x27;
    temp2.addr[1] = 0x2e;
    temp2.addr[0] = 0x84;
    
    
    temp2.addr_type = BLE_GAP_ADDR_TYPE_RANDOM_STATIC;

    ble_gap_addr_t       *whitelist_addr[1];

    /* Setup whitelist */
    whitelist_addr[0] = &temp2;
    whitelist.pp_addrs = whitelist_addr;
    whitelist.addr_count = 1;
    whitelist.pp_irks = NULL;
    whitelist.irk_count =0;

    adv_params.type = BLE_GAP_ADV_TYPE_ADV_IND;
    adv_params.p_peer_addr = NULL;
    adv_params.fp = BLE_GAP_ADV_FP_FILTER_CONNREQ;
    adv_params.interval = 64;
    adv_params.timeout = 180;
    adv_params.p_whitelist = &whitelist;

   advertising_init(BLE_GAP_ADV_FLAG_BR_EDR_NOT_SUPPORTED);

    // Start advertising.
    err_code = sd_ble_gap_adv_start(&adv_params);
    APP_ERROR_CHECK(err_code);
}


static void on_ble_evt(ble_evt_t * p_ble_evt)
{
    uint32_t                         err_code;
    ble_gap_enc_info_t *             p_enc_info;
    static ble_gap_sec_keyset_t      s_sec_keyset;

    switch (p_ble_evt->header.evt_id) {
    case BLE_GAP_EVT_CONNECTED:
        status.m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
        status.system_status = BLE_CONNECTED;
        break;

    case BLE_GAP_EVT_DISCONNECTED:
        status.m_conn_handle = BLE_CONN_HANDLE_INVALID;
        advertising_start();
        break;

    case BLE_EVT_TX_COMPLETE:
        break;

    case BLE_GAP_EVT_SEC_PARAMS_REQUEST:
        s_sec_keyset.keys_central.p_enc_key = NULL;
        s_sec_keyset.keys_central.p_id_key = NULL;
        s_sec_keyset.keys_central.p_sign_key = NULL;
        err_code = sd_ble_gap_sec_params_reply(status.m_conn_handle,
            BLE_GAP_SEC_STATUS_SUCCESS,
            &m_sec_params,
            &s_sec_keyset);
        APP_ERROR_CHECK(err_code);
        break;

    case BLE_GATTS_EVT_SYS_ATTR_MISSING:
        err_code = sd_ble_gatts_sys_attr_set(status.m_conn_handle, NULL, 0, 0);
        APP_ERROR_CHECK(err_code);
        break;

    case BLE_GAP_EVT_AUTH_STATUS:
        break;

    case BLE_GAP_EVT_SEC_INFO_REQUEST:
        if (s_sec_keyset.keys_periph.p_enc_key != NULL) {
            p_enc_info = &s_sec_keyset.keys_periph.p_enc_key->enc_info;

            err_code = sd_ble_gap_sec_info_reply(status.m_conn_handle, p_enc_info, NULL, NULL);
            APP_ERROR_CHECK(err_code);
        }
        else {
            // No keys found for this device.
            err_code = sd_ble_gap_sec_info_reply(status.m_conn_handle, NULL, NULL, NULL);
            APP_ERROR_CHECK(err_code);
        }
        break;

        // indication
    case BLE_GATTS_EVT_HVC:
#ifdef FLASH_FEATURE
        flash_context.indication_pending = false;
#endif
        break;

    case BLE_GAP_EVT_TIMEOUT:
        //should not happen (BLE Timeout desactivated)
        break;

    default:
        // No implementation needed.
        break;
    }
}


static void ble_evt_dispatch(ble_evt_t * p_ble_evt)
{
    ble_conn_params_on_ble_evt(p_ble_evt);
#ifdef BLE_DFU_APP_SUPPORT
    /** @snippet [Propagating BLE Stack events to DFU Service] */
    ble_dfu_on_ble_evt(&m_dfus, p_ble_evt);
    /** @snippet [Propagating BLE Stack events to DFU Service] */
#endif // BLE_DFU_APP_SUPPORT
    on_ble_evt(p_ble_evt);
}


static void sys_evt_dispatch(uint32_t sys_evt)
{
    pstorage_sys_event_handler(sys_evt);
}


static void ble_stack_init(void)
{
    uint32_t err_code;

    // Initialize the SoftDevice handler module.
    //    SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_XTAL_20_PPM, false);
    SOFTDEVICE_HANDLER_INIT(NRF_CLOCK_LFCLKSRC_RC_250_PPM_4000MS_CALIBRATION, false);

    // Enable BLE stack
    ble_enable_params_t ble_enable_params;
    memset(&ble_enable_params, 0, sizeof(ble_enable_params));
    ble_enable_params.gatts_enable_params.service_changed = IS_SRVC_CHANGED_CHARACT_PRESENT;
    err_code = sd_ble_enable(&ble_enable_params);
    APP_ERROR_CHECK(err_code);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_ble_evt_handler_set(ble_evt_dispatch);
    APP_ERROR_CHECK(err_code);

    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}


static void init_sys_evt_handler() {
    uint32_t err_code;
    // Register with the SoftDevice handler module for BLE events.
    err_code = softdevice_sys_evt_handler_set(sys_evt_dispatch);
    APP_ERROR_CHECK(err_code);
}


static void scheduler_init(void)
{
    APP_SCHED_INIT(SCHED_MAX_EVENT_DATA_SIZE, SCHED_QUEUE_SIZE);
}


static void manage_adv(void)
{
    uint32_t err_code;

    if ((status.jobtodo & STOP_ADV) == STOP_ADV) {
        status.jobtodo -= STOP_ADV;
        err_code = sd_ble_gap_adv_stop();           // stop advertising.
        APP_ERROR_CHECK(err_code);
        status.system_status = BLE_OFF;
    }
    else if ((status.jobtodo & RESTART_ADV) == RESTART_ADV) {
        status.jobtodo -= RESTART_ADV;
        advertising_start();                        // restart advertising.
    }
}



void main_loop(void)
{
    for (;;) {
        app_sched_execute();
        manage_adv();
    }
}


void wdt_init(void)
{
    NRF_WDT->CONFIG = (WDT_CONFIG_HALT_Pause << WDT_CONFIG_HALT_Pos) | (WDT_CONFIG_SLEEP_Run << WDT_CONFIG_SLEEP_Pos);   //Configure Watchdog. a) Pause watchdog while the CPU is halted by the debugger.  b) Keep the watchdog running while the CPU is sleeping.
    NRF_WDT->CRV = 30 * 32768;
    NRF_WDT->RREN |= WDT_RREN_RR0_Msk;  //Enable reload register 0
    NRF_WDT->TASKS_START = 1;           //Start the Watchdog timer
}


int main(void)
{
    timers_init();
    ble_stack_init();
    scheduler_init();
    twi_master_init();
    wdt_init();
    gap_params_init();
    services_init();
    advertising_init(BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE);
    conn_params_init();
    sec_params_init();
    advertising_start();
    main_loop();

    return 0;
}
